# - coding: utf-8 -

import time
import datetime
###############

import telepot
import roller


from utils import dump, save_stats
from archive import Archive
#####################################################
from errors import RollStringError, RollTooManyDice



###############################################
SHELVE_FILE = "db_general.db"

TIME_LIMIT = 60 * 30 #time in seconds # half an hour

#tuple of admin user
ADMIN_USERS = (
	108066976,
)

#tuple of banned id
BANNED_ID = (
	0,

)


##############################

REQUEST_COUNT = {}
REQUEST_TIME_LIMIT = 60 * 30 #time in seconds #half an hour
REQUEST_COUNT_LIMIT = 50
REQUEST_WARNING_LIMIT = 5


##########################################
HELP_DICT = {}
HELP_TIME_LIMIT = 60 * 60 # 1 hour
HELP_WARNING_LIMIT = 5

###################################

GLOBAL_DATA = Archive(SHELVE_FILE)
GLOBAL_DATA["logs"] = GLOBAL_DATA.data.get("logs", [])
GLOBAL_DATA["banned_id"] = GLOBAL_DATA.data.get("banned_id", [])
GLOBAL_DATA["admin_id"] = GLOBAL_DATA.data.get("admin_id", [108066976])

GLOBAL_DATA["saved_rolls"] = GLOBAL_DATA.data.get("saved_rolls", {}) #id user -> type roll -> name roll -> string roll


# GLOBAL_DATA["saved_rolls"]["108066976"] = {
# 	"roll" : {
# 		"dd_chargen" : "3/4d6",
# 		"dardo_incantato" : "d6+1"
# 	}
# }

#BANNED_ID = GLOBAL_DATA["banned_id"]

################################################################
################################################################



def log_error(msg, error, add_msg=""):
	print(">" * 80)
	
	err_msg = "{}: {}".format(error.code, error.message)
	if add_msg:
		err_msg += " => " + add_msg
	
	print(err_msg)
	
	if error.data:
		print("*" * 80)
		dump(error.data)
		print("*" * 80)

	print("<" * 80)

	return True


def get_name_from(msg):
	if msg["from"].get("username"):
		return "@" + msg["from"]["username"]
	
	name_from = msg["from"]["first_name"]

	if msg["from"].get("last_name"):
		return name_from + " " + msg["from"]["last_name"]
	
	return name_from



def help_handler(chat_id, msg, bot):
	
	chat_from = str(chat_id)
	
	if chat_from not in HELP_DICT:
		HELP_DICT[chat_from] = {"warnings" : HELP_WARNING_LIMIT}
		HELP_DICT[chat_from]["ts_end"] = (time.time() + HELP_TIME_LIMIT)

	
	if HELP_DICT[chat_from]["ts_end"] < time.time():
		HELP_DICT[chat_from]["warnings"] = HELP_WARNING_LIMIT
		HELP_DICT[chat_from]["ts_end"] = (time.time() + HELP_TIME_LIMIT)
	

	if HELP_DICT[chat_from]["warnings"]:
		HELP_DICT[chat_from]["warnings"] -= 1
		bot.sendMessage(chat_id,
			"\n".join((
				"Wrong syntax!",
				"""use "/roll dicetype" or "/pool dicetype" """,
				"Examples:",
				"/roll 3d6",
				"/pool 5d10",
				"Info at: /help"
				)),
			"HTML",
			True, #no preview
			True, #no sound
			msg["message_id"])

	return True


def show_help(chat_id, msg, bot):
	
	
	bot.sendMessage(chat_id,
		"""
		/roll x/yDn+z[+...] => (x best of y dice of n face)
			Es: 3/4D6 (best 3 dice of 4D6), 2D8+1d6, 1D100, ...

		/pool xDn => pools x dice with n face. add <code>(t)</code> to calculate the threshold results
			Es: 5d10(8)... 6d6(-3)
			Negative threshold are used to find value or less.
			Can also separate high results and low results:
				5d10(8,10,1) ... 6d6(,5,2)

		/help this info
		""",
	"HTML", True, True) #format html, disabled page_preview, disable notification 

	
	return True


def save_user_roll(section, chat_id, msg, bot):
	
	text_raw = " ".join(msg["text"].split()[1:])
	(save_name, roll_string) = text_raw.split(" ", 1)

	id_user = str(msg["from"].get("id"))
	
	if id_user not in GLOBAL_DATA["saved_rolls"]:
		GLOBAL_DATA["saved_rolls"][id_user] = {}

	if section not in GLOBAL_DATA["saved_rolls"][id_user]:
		GLOBAL_DATA["saved_rolls"][id_user][section] = {}

	GLOBAL_DATA["saved_rolls"][id_user][section][save_name] = roll_string.replace(" ", "") #normalized roll_string

	Archive.save()

	if section == "roll":
		return roll_values(chat_id, msg, bot)

	if section == "pool":
		return pool_values(chat_id, msg, bot)


def roll_values(chat_id, msg, bot):
	id_user = str(msg["from"].get("id"))

	try:
		search_saved = msg["text"].split()[1]
		roll_string = GLOBAL_DATA["saved_rolls"][id_user]["roll"][search_saved]
		roll_loaded = True
	except:
		roll_loaded = False
		text_raw = " ".join(msg["text"].split()[1:])
		roll_string = text_raw.replace(" ", "") #normalized roll_string

	if not roll_string.strip():
		
		# if msg["from"].get("id") not in ADMIN_USERS:
		# 	print("NO ROLL STRING")
		# 	return help_handler(chat_id, msg, bot)
		# 	return True

		print("NO ROLL STRING")
		return help_handler(chat_id, msg, bot)
		return True

		# kblist = []
		# kb = [["d4", "d6", "d8"], ["d10", "d12", "d20"]]
		# 
		# for i, buttons in enumerate(kb):
		# 	kblist.append([])
		# 	
		# 	for txt in buttons: 
		# 		kblist[i].append({
		# 			"text" : txt,
		# 			"callback_data" : {
		# 				"id" : "{}_{}".format(msg["message_id"], txt),
		# 				"from" : msg["from"],
		# 				"message" : "/roll " + txt
		# 			}
		# 		})
		# 
		# 
		# bot.sendMessage(108066976, "blaaah!", "HTML", True, True, msg["message_id"], {"keyboard" : kblist, "one_time_keyboard":True} )
		# 
		# return True		
		
	#log_data = {"msg" : msg}
	#GLOBAL_DATA["logs"].append(log_data)

	try:
		(result, rolls) = roller.multi_roll(roll_string)
		#log_data["results"] = (result, rolls)
	
	except RollStringError as e:
		return log_error(msg, e, roll_string)
	
	except RollTooManyDice as e:
		return log_error(msg, e, roll_string)

	
	results_tmpl = "{}: {}"
	rolls_result = [results_tmpl.format(r[3], r[1]) for r in rolls if "d" in r[3].lower()]

	if roll_loaded:
		response_result = "Result: <b>{}</b> (from <em>{}</em>)\n".format(result, roll_string)
	else:
		response_result = "Result: <b>{}</b>\n".format(result)

	response_html = response_result + "Values => {}".format(
		", ".join(rolls_result)
	)

	bot.sendMessage(chat_id, response_html,
		"HTML", #format ah HTML
		True, #disable page preview
		True, #disable sound notification
		msg["message_id"]
	)

	print(response_html); save_stats(msg)
	#GLOBAL_DATA.save()

		
	return True	



def pool_values(chat_id, msg, bot):

	id_user = str(msg["from"].get("id"))

	try:
		search_saved = msg["text"].split()[1]
		roll_string = GLOBAL_DATA["saved_rolls"][id_user]["pool"][search_saved]
		roll_loaded = True
	except:
		roll_loaded = False
		text_raw = " ".join(msg["text"].split()[1:])
		roll_string = text_raw.replace(" ", "") #normalized roll_string


	if not roll_string.strip():
		print("NO POOL STRING")
		return help_handler(chat_id, msg, bot)
		return True


	#log_data = {"msg" : msg}
	#GLOBAL_DATA["logs"].append(log_data)

	try:
		(result, rolls, res_calculated, high, low) = roller.pool_roll(roll_string)
		#log_data["results"] = (result, rolls, res_calculated)

	except RollStringError as e:
		return log_error(msg, e, roll_string)
	
	except RollTooManyDice as e:
		return log_error(msg, e, roll_string)
	

	result_str = ""
	if res_calculated:
		if roll_loaded:
			result_str = "Result: <b>{}</b> (from <em>{}</em>)\n".format(result, roll_string)
		else:
			result_str = "Result: <b>{}</b>\n".format(result)

	response = "{}Values => {}".format(
		result_str,
		rolls
	)
	
	if res_calculated or high or low:
		
		result_str = val_h = val_low = ""

		if high:
			val_h = len(rolls[0])

		if low:
			val_low = len(rolls[-1])
			
		res_h_l = []

		if high:
			res_h_l.append("\nHigh ({}+): {}".format(high, val_h))

		if low:
			res_h_l.append("\nLow ({}-): {}".format(low, val_low))
		
		res_h_l = "".join(res_h_l)
		
		
		response += res_h_l 
		

	bot.sendMessage(chat_id, response,
		"HTML", #format ah HTML
		True, #disable page preview
		True, #disable sound notification
		msg["message_id"]
	)

	print(response); save_stats(msg)

	#GLOBAL_DATA.save()

	return True	


def admin_management(chat_id, msg, bot):
	from_id = msg.get("from", {}).get("id")

	if from_id not in ADMIN_USERS:
		return True

	commands_raw = msg["text"].lower().split()[1:]
	commands = msg["text"].split()[1:]
	
	if commands_raw[1] == "ban":
		user_id = commands[2]

		GLOBAL_DATA["banned_id"].append(user_id)
		GLOBAL_DATA.save()
		
		bot.sendMessage(chat_id,"User <b>{}</b> banned from bot".format(user_id), "HTML", True, True)

		return True		

	return True



def main_handler(msg, bot=None):
	
	from_id = msg.get("from", {}).get("id")
	
	if from_id in BANNED_ID:
		print("")
		print("USER BANNED:")
		dump(msg.get("from"))
		print("")
		return True
	
	if not msg.get("date"):
		print("MESSAGGIO NON DATATO")
		return True

	msg[u"date_str"] = datetime.datetime.fromtimestamp(msg["date"]).strftime("%Y-%m-%d, %H:%M:%S")

	print("=" * 80)
	dump(msg)
	print("=" * 80)


	if (msg["date"] + TIME_LIMIT) < time.time():
		print("**** MESSAGGIO TROPPO VECCHIO ****")
		return True


	(content_type, chat_type, chat_id) = telepot.glance(msg)

	if from_id == 155230170:
	#	bot.sendMessage(chat_id,"Ho diciuto 'Basta spammare, Terry!'", "HTML", True, True)
	#	return True
		pass


	if content_type != "text":
		return True

	#if not msg["message_id"] > bot.manager.last_id:
	#	print("MESSAGGIO VECCHIO")
	#	return True
	
	bot.manager.last_id = msg["message_id"]
	
	txt_check = msg["text"].lower()


	chat_from = str(chat_id)

	try:
		REQUEST_COUNT[chat_from]["count"] += 1

	except:
		REQUEST_COUNT[chat_from] = {"count" : 1}
		REQUEST_COUNT[chat_from]["ts_end"] = (time.time() + REQUEST_TIME_LIMIT)
		REQUEST_COUNT[chat_from]["warnings"] = REQUEST_WARNING_LIMIT
		REQUEST_COUNT[chat_from]["date_end"] = datetime.datetime.fromtimestamp(REQUEST_COUNT[chat_from]["ts_end"]).strftime("%A %Y-%m-%d, %H:%M:%S")

	
	chat_req_d = REQUEST_COUNT[chat_from]
	
	time_limit = (time.time() - REQUEST_TIME_LIMIT)
	
	if chat_req_d["ts_end"] < time.time():
		chat_req_d["count"] = 1
		chat_req_d["ts_end"] = (time.time() + REQUEST_TIME_LIMIT)
		chat_req_d["warnings"] = REQUEST_WARNING_LIMIT
		chat_req_d["date_end"] = datetime.datetime.fromtimestamp(chat_req_d["ts_end"]).strftime("%A %Y-%m-%d, %H:%M:%S")

	if chat_req_d["count"] > REQUEST_COUNT_LIMIT:
		if chat_req_d["warnings"]:
			chat_req_d["warnings"] -= 1
			bot.sendMessage(chat_id,
				"\n".join((
					"DiceRollerBot has a limit of {} requests every {} seconds on this chat.".format(
																		REQUEST_COUNT_LIMIT, REQUEST_TIME_LIMIT),
					"Please wait until {}".format(chat_req_d["date_end"]),
					"(Warnings left: {})".format(chat_req_d["warning"])
						
					)),
				"HTML",
				True, #no preview
				True, #no sound
				msg["message_id"])

		return True


	command = txt_check.split(" ", 1)[0]
	command = command.split("@", 1)[0]

	#execute action on valid command
	if command == "/start":
		return show_help(chat_id, msg, bot)

	if command == "/rollsave":
		return save_user_roll("roll", chat_id, msg, bot)

	if command == "/poolsave":
		return save_user_roll("pool", chat_id, msg, bot)

	if command == "/roll":
		return roll_values(chat_id, msg, bot)

	if command == "/pool":
		return pool_values(chat_id, msg, bot)

	if command == "/help":
		return show_help(chat_id, msg, bot)

	if command == "/admin":
		return admin_management(chat_id, msg, bot)
	
	if command == "/stop":
		return True

	print("NO VALID COMMAND")

	return help_handler(chat_id, msg, bot)

	return True
	

		 
